//======================================
//FilaName	=	sample.c
//LastDate	=	20250409
//======================================
//实际测试OK!

//功能说明
//本例程基于STC32G核心转接板（屠龙刀）进行编写测试。
//使用Keil C251编译器，Memory Model推荐设置XSmall模式，默认定义变量在edata，单时钟存取访问速度快。
//edata建议保留1K给堆栈使用，空间不够时可将大数组、不常用变量加xdata关键字定义到xdata空间。
//I2C总线传感器DHT12的温湿度数据
//串口(115200,N,8,1)打印采集结果
//下载时, 默认时钟 12MHz (影响延时函数!!!).
//

#include <STC32G.h>
#include <intrins.h>
#include <stdio.h>

#define MAIN_Fosc		12000000UL


#define Baudrate		115200L
#define TM				(65536 -(MAIN_Fosc/Baudrate/4))
#define PrintUart		1			//1:printf 使用 UART1; 2:printf 使用 UART2

sbit DHT_SDA = P1^6;				//DHT12模块的SDA引脚,51准双向接口
sbit DHT_SCL = P1^7;				//DHT12模块的SCL引脚,51准双向接口

float Temprature;
float Humi;
unsigned char Sensor_AnswerFlag;
unsigned char Sensor_ErrorFlag;

//unsigned int Temperature[2];
//unsigned int Humidity[2];

//unsigned char Data_Receive[5];

void delay_ms(unsigned int ms);
unsigned char DHT12_READ(void);

//串口打印函数
void UartInit(void)
{
#if(PrintUart == 1)
	SCON = (SCON & 0x3f) | 0x40; 
	T1x12 = 1;				//定时器时钟1T模式
	S1BRT = 0;          	//串口1选择定时器1为波特率发生器
	TL1  = TM;
	TH1  = TM>>8;
	TR1 = 1;				//定时器1开始计时

//	SCON = (SCON & 0x3f) | 0x40; 
//	T2L  = TM;
//	T2H  = TM>>8;
//	AUXR |= 0x15;   //串口1选择定时器2为波特率发生器
#else
	S2_S = 1;       //UART2 switch to: 0: P1.0 P1.1,  1: P4.6 P4.7
    S2CFG |= 0x01;  //使用串口2时，W1位必需设置为1，否则可能会产生不可预期的错误
	S2CON = (S2CON & 0x3f) | 0x40; 
	T2L  = TM;
	T2H  = TM>>8;
	AUXR |= 0x14;	      //定时器2时钟1T模式,开始计时
#endif
}

void UartPutc(unsigned char dat)
{
#if(PrintUart == 1)
	SBUF = dat; 
	while(TI==0);
	TI = 0;
#else
	S2BUF  = dat; 
	while(S2TI == 0);
	S2TI = 0;    //Clear Tx flag
#endif
}

char putchar(char c)
{
	UartPutc(c);
	return c;
}


void main(void)
{
	WTST = 0;  //设置程序指令延时参数，赋值为0可将CPU执行指令的速度设置为最快
	EAXFR = 1; //扩展寄存器(XFR)访问使能
	CKCON = 0; //提高访问XRAM速度
	
	P0M1 = 0x00;   P0M0 = 0x00;   		//设置为准双向口
	P1M1 = 0x00;   P1M0 = 0x00;   		//设置为准双向口
	P2M1 = 0x00;   P2M0 = 0x00;   		//设置为准双向口
	P3M1 = 0x00;   P3M0 = 0x00;   		//设置为准双向口
	P4M1 = 0x00;   P4M0 = 0x00;   		//设置为准双向口
	P5M1 = 0x00;   P5M0 = 0x00;   		//设置为准双向口
	P6M1 = 0x00;   P6M0 = 0x00;   		//设置为准双向口
	P7M1 = 0x00;   P7M0 = 0x00;   		//设置为准双向口
	
	UartInit();
	
	//DHT12 上电后要等待 1S 以越过不稳定状态在此期间不能发送任何指令
	DHT_SDA =0;
	DHT_SCL =0;
	
	delay_ms(1200);
	printf("STC32G DHT12 Test.\r\n");
	
	while(1)
	{
	    DHT12_READ();
	    delay_ms(1000);
	}
}

//由于延时函数使用指令循环,所以不仅注意时钟使用12MHz,而且编译优化级别也必须使用默认的7级优化!
void Delay30us(void)//@12.000MHz
{
unsigned long i;

	_nop_();
	_nop_();
	_nop_();
	i = 88UL;
	while(i)i--;
}

void Delay20ms(void)//@12.000MHz
{
unsigned long i;

	_nop_();
	_nop_();
	_nop_();
	i = 59998UL;
	while(i)i--;
}

void delay_ms(unsigned int ms)
{
unsigned int i;
	do{
	    i = MAIN_Fosc / 6000;
	    while(--i);   					//6T per loop
	}while(--ms);
}


//读DHT12的一个字节
unsigned char DHT12_Rdata(void)
{
unsigned char	i;
unsigned int	j;
unsigned char	dat =0, rbit =0;

	for(i=0; i<8; i++)
	{
		DHT_SDA =1;		//读数据前先输出1
		while(!DHT_SDA)	//检测上次低电平是否结束
		{
			if(++j >=20000)
			{
				break;	//防止死循环
			}
		}
		
		//延时 26~70us, 跳过数据'0'的高电平
		Delay30us();
		
		//判断DHT12的数据位
		rbit =0;
		if(DHT_SDA) rbit=1;
		j=0;
		while(DHT_SDA)
		{
			if(++j >=20000) break;
		}
		
		dat <<=1;
		dat |= rbit;
	}

	return(dat);
}


//读取DHT12的温湿度数据
unsigned char	DHT12_READ(void)
{
unsigned long  j;
unsigned char	Humi_H, Humi_L, Temp_H, Temp_L, Temp_CAL, temp;

	DHT_SDA =0;
	Delay20ms();
	DHT_SDA =1;
	Delay30us();
	
	Sensor_AnswerFlag =0;
	if(!DHT_SDA)
	{
		Sensor_AnswerFlag =1;	//收到起始信号
		
		j =0;
		while(!DHT_SDA)
		{
			if(++j >=500)
			{
				Sensor_ErrorFlag =1;
				break;
			}
		}
		
		j=0;
		while(DHT_SDA)	//判断是否有80us的高电平
		{
			if(++j >800)
			{
				Sensor_ErrorFlag =1;
				break;
			}
		}
		
		//接收数据
		Humi_H = DHT12_Rdata();
		Humi_L = DHT12_Rdata();
		Temp_H = DHT12_Rdata();
		Temp_L = DHT12_Rdata();
		Temp_CAL = DHT12_Rdata();
		
		temp = (unsigned char)(Humi_H +Humi_L +Temp_H +Temp_L);	//只取和的低字节
		
		if(temp == Temp_CAL)
		{	//校验正确
			Humi = Humi_H *10 +Humi_L;
			
			if(Temp_L &0x80)
			{	//温度数据为负数
				Temprature = 0 - (Temp_H *10 + (Temp_L &0x7f) );
			}
			else
			{	//温度数据为正数
				Temprature = Temp_H *10 + (Temp_L &0x7f);
			}
			
			//判断数据是否超范围: 温度=-20~+60摄氏度, 湿度 20%~95% RH;
			if(Humi >950) Humi =950;
			if(Humi <200) Humi =200;
				
			if(Temprature >600) Temprature =600;
			if(Temprature <-200) Temprature =-200;
				
			Temprature /= 10;	//得到最后温度值
			Humi /= 10;			//得到最后湿度值
			
			printf("\r\n T= %.1f \r\n", Temprature);
			printf("Humi = %.1f %%RH\r\n", Humi);
		}
		else
		{
			printf("CAL Error!!\r\n");		
			printf("%d \r%d \r%d \r%d \r%d \r%d \r\n", Humi_H, Humi_L, Temp_H, Temp_L, Temp_CAL, temp);
		}
	}
	else
	{
		Sensor_ErrorFlag = 0;	//未收到传感器的响应
		printf("Sensor Error!!\r\n");
	}
	
	return 1;
}


//End Of File
